/* Standard C includes */

#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stddef.h>
#include <stdarg.h>

/* POSIX includes */

#include <pthread.h>
#include <unistd.h>

/* Socket includes */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

/* Local includes */

#include "ccbcontrolclient.h"
#include "ccbclientpriv.h"
#include "netbuf.h"
#include "ccbnetobj.h"
#include "ccbconstants.h"
#include "ccbqueue.h"
#include "ccbcontrolmsg.h"
#include "tcpip.h"

/*-----------------------------------------------------------------------*
 * The functions in this file handle the client end of the control link	 *
 * to a remote CCB server.						 *
 *									 *
 * Conventions:								 *
 *  All external functions are named starting with the prefix ccb_cc_.	 *
 *  These functions must check their arguments, and surround all	 *
 *  accesses to their CCBControlClient argument with calls to		 *
 *  cc_lock_client() and cc_unlock_client(), to give them and any	 *
 *  internal functions that they call, exclusive access to the contents	 *
 *  of said object.							 *
 *									 *
 *  All internal functions are named starting with the prefix cc_.	 *
 *  These functions all assume that their caller has acquired		 *
 *  exclusive access to the CCBControlClient object that they are	 *
 *  passed, and are careful to always reacquire exclusive access	 *
 *  before returning to the caller, if they ever relinquish it.		 *
 *									 *
 *  To avoid non-portable recursive locking of the guard mutex,		 *
 *  internal functions temporarily relinquish exclusive access to	 *
 *  their CCBControlClient object before calling manager callback	 *
 *  functions. Since ccb_log_client_msg() calls a callback function	 *
 *  of the parent CCBClientLink, it is essential to call			 *
 *  cc_error_while_locked(), instead of the above function when		 *
 *  reporting an error message while hodling exclusive access to	 *
 *  a CCBControlClient object. This function relinquishes exclusive	 *
 *  access for the duration of the call to ccb_log_client_msg().	 *
 *									 *
 *  Functions that don't use their return values for anything else,	 *
 *  return an integer which is 1 if an error occurred, or 0 if the	 *
 *  function was successful.						 *
 *									 *
 *  Error conditions are reported once, as close as possible to the	 *
 *  site of the error. Higher level functions, on receiving an error	 *
 *  return from an internal function, assume that the error has been	 *
 *  reported by that function, or by a function that it called. Since	 *
 *  low-level external utility functions don't report log errors, it	 *
 *  is up to the function in this module that calls one of these	 *
 *  functions to report the error.					 *
 *-----------------------------------------------------------------------*/

/*
 * Collect all callback functions and callback data.
 */
typedef struct {
  CCBStatusReplyFn *status_reply_fn;/* The status-request reply */
                                             /*  callback. */
  void *status_reply_data;              /* status_reply_fn() */
                                             /*  callback data. */
  CCBCmdErrorFn *cmd_error_fn;               /* The callback to use to */
                                             /*  report control command */
                                             /*  errors. */
  void *cmd_error_data;                      /* The callback data of */
                                             /*  cmd_error_fn() */
} CCBControlClientCallbacks;

/*
 * The following constant is designed to be assigned to any
 * CCBControlClientCallbacks structure to set all of its contained
 * function pointers and data pointers to their respective versions of
 * nul. This takes advantage of the fact that the C standard states
 * that statically allocated structures are initialized as though all
 * of their members were individually assigned 0 (NB this isn't the
 * same as setting all bytes to zero when pointers are involved).  Not
 * only does this tremendously reduce the number of statements needed
 * to initialize one of these structures in the CCBControlClient
 * constructor, but it is also safer, because it makes it impossible
 * to forget to initialize a newly added callback and its data
 * pointer.
 */
static const CCBControlClientCallbacks init_client_callbacks;

/*
 * Enumerate the states that the control link goes through.
 */
typedef enum {
  CC_LINK_DOWN,          /* The control link isn't connected */
  CC_LINK_CONNECTING,    /* The control link is being established */
  CC_LINK_UP,            /* The control link is connected and usable */
  CC_LINK_DISCONNECTING  /* The control link is being disconnected */
} CCLinkState;

/*
 * Define the object that manages the CCB server end of the control link.
 */
struct CCBControlClient {
  pthread_mutex_t guard;        /* The mutex which controls access to the */
                                /*  other members of this object. */
  int guard_ready;              /* True after 'guard' has been allocated */
  pthread_cond_t herald;        /* The condition variable that signals */
                                /*  changes in I/O status to readers and */
                                /*  writers, when using blocking I/O. */
  int herald_ready;             /* True, once 'herald' has been initialized */
  int read_locked;              /* True while a thread has been granted */
                                /*  exclusive control-socket read-access */
  int write_locked;             /* True while a thread has been granted */
                                /*  exclusive control-socket write-access */
  int write_busy;               /* True while a message is in the process */
                                /*  of being written to the control socket. */
  int blocking;                 /* True when using blocking socket I/O */
  int sending_shutdown;         /* A shutdown or reboot command is in the */
                                /*  process of being sent. When it has been */
                                /*  sent, the connection will be terminated. */
  CCLinkState link_state;       /* The internal status of the control link */
  CCBControlClientCallbacks cb; /* Callback functions */
  CCBClientLink *client;            /* The parent CCB client */
  NetSendStr *nss;              /* The network output stream */
  NetReadStr *nrs;              /* The network input stream */
  CCBMessageQueue *mq;          /* The queue of outgoing messages */
  CCBConfig *cnf;               /* The configuration of the most recently */
                                /*  started scan or intra-scan. */
};

static int ccb_cc_queue_control_command(CCBControlClient *cc,
					CCBControlCommandHeader *head,
					CCBControlCommandType type,
					long id);
static int cc_deliver_reply(CCBControlClient *cc, CCBControlReply *reply);
static int cc_stage_next_command(CCBControlClient *cc);
static int cc_queue_control_command(CCBControlClient *cc,
				    CCBControlCommandHeader *head);
static int cc_unqueue_next_command(CCBControlClient *cc,
				   CCBControlCommand *msg);
static int cc_send_control(CCBControlClient *cc);
static int cc_read_control(CCBControlClient *cc);
static int cc_command_error(CCBControlClient *cc, long id,
			    CCBCmdStatus status);
static int cc_socket_fd(CCBControlClient *cc);
static int cc_connect_link(CCBControlClient *cc, const char *host,
			   int nonblocking);
static void cc_disconnect_link(CCBControlClient *cc);
static int cc_connection_completed(CCBControlClient *cc, int sock);
static int cc_activate_link(CCBControlClient *cc);
static int cc_lock_client(CCBControlClient *cc);
static int cc_unlock_client(CCBControlClient *cc);
static int cc_error_while_locked(CCBControlClient *cc, CCBLogLevel level,
				 unsigned long id, const char *fmt, ...);
static int cc_queue_config_updates(CCBControlClient *cc, long id,
				   const CCBConfig *cnf);

/*.......................................................................
 * Create a new CCBControlClient object.
 *
 * Input:
 *  client         CCBClientLink *  The parent CCB client.
 * Output:
 *  return  CCBControlClient *  The new object, or NULL on error.
 */
CCBControlClient *new_CCBControlClient(CCBClientLink *client)
{
  CCBControlClient *cc;  /* The object to be returned */
  char errmsg[200];      /* An error message */
  int status;            /* The return value of a POSIX function */
/*
 * Check the arguments.
 */
  if(!client) {
    ccb_log_client_msg(NULL, CCB_ERROR, CCB_LOGID(160),
		"new_CCBControlClient: NULL argument(s).\n");
    return NULL;
  };
/*
 * Check that the tables of control message descriptions are valid.
 */
  if(ccb_check_net_table(&ccb_control_cmd_table, errmsg,
			 sizeof(errmsg))) {
    ccb_log_client_msg(NULL, CCB_ERROR, CCB_LOGID(161),
		       "Control command-table: %s\n", errmsg);
    return NULL;
  };
  if(ccb_check_net_table(&ccb_control_reply_table, errmsg,
			 sizeof(errmsg))) {
    ccb_log_client_msg(NULL, CCB_ERROR, CCB_LOGID(162),
		       "Control reply-table: %s\n", errmsg);
    return NULL;
  };
/*
 * Allocate the container.
 */
  cc = malloc(sizeof(CCBControlClient));
  if(!cc) {
    ccb_log_client_msg(NULL, CCB_ERROR, CCB_LOGID(163),
		"new_CCBControlClient: Insufficient memory.\n");
    return NULL;
  };
/*
 * Before attempting any operation that might fail, initialize the
 * container at least up to the point at which it can safely be passed
 * to del_CCBControlClient().
 */
  cc->guard_ready = 0;
  cc->herald_ready = 0;
  cc->read_locked = 0;
  cc->write_locked = 0;
  cc->write_busy = 0;
  cc->blocking = 1;
  cc->sending_shutdown = 0;
  cc->link_state = CC_LINK_DOWN;
  cc->cb = init_client_callbacks;
  cc->client = client;
  cc->nss = NULL;
  cc->nrs = NULL;
  cc->mq = NULL;
  cc->cnf = NULL;
/*
 * Get a CCB configuration object, initialized with the power-on
 * default configuration of the CCB.
 */
  cc->cnf = new_CCBConfig();
  if(!cc->cnf)
    return del_CCBControlClient(cc);
/*
 * Allocate the mutex which is used to sequence accesses to the new object.
 */
  status = pthread_mutex_init(&cc->guard, NULL);
  if(status) {
    ccb_log_client_msg(client, CCB_ERROR, CCB_LOGID(164),
	"new_CCBControlClient: pthread_mutex_init -> %s\n", strerror(status));
    return del_CCBControlClient(cc);
  };
  cc->guard_ready = 1;
/*
 * Initialize the condition variable used to signal changes in I/O status.
 */
  status = pthread_cond_init(&cc->herald, NULL);
  if(status) {
    ccb_log_client_msg(client, CCB_ERROR, CCB_LOGID(165),
	"new_CCBControlClient: pthread_cond_init -> %s\n", strerror(status));
    return del_CCBControlClient(cc);
  };
  cc->herald_ready = 1;
/*
 * Allocate the output stream manager.
 */
  cc->nss = new_NetSendStr(-1, ccb_net_max_obj_size(&ccb_control_cmd_table));
  if(!cc->nss) {
    ccb_log_client_msg(client, CCB_ERROR, CCB_LOGID(166),
	"new_CCBControlClient: Insufficient memory for output stream.\n");
    return del_CCBControlClient(cc);
  };
/*
 * Allocate the input stream manager.
 */
  cc->nrs = new_NetReadStr(-1, ccb_net_max_obj_size(&ccb_control_reply_table));
  if(!cc->nrs) {
    ccb_log_client_msg(client, CCB_ERROR, CCB_LOGID(167),
	"new_CCBControlClient: Insufficient memory for input stream.\n");
    return del_CCBControlClient(cc);
  };
/*
 * Allocate the message queue that is used to queue outgoing control reply
 * messages.
 */
  cc->mq = new_CCBMessageQueue(sizeof(CCBControlCommand), CCB_CNTRL_CMD_QLEN,
			       CCB_MQ_DROP);
  if(!cc->mq) {
    ccb_log_client_msg(client, CCB_ERROR, CCB_LOGID(168),
	"new_CCBControlClient: Insufficient memory for message queue.\n");
    return del_CCBControlClient(cc);
  };
  return cc;
}

/*.......................................................................
 * Disconnect and delete a CCBControlClient object.
 *
 * Input:
 *  cc     CCBControlClient *  The object to be deleted.
 * Output:
 *  return CCBControlClient *  The deleted object (always NULL).
 */
CCBControlClient *del_CCBControlClient(CCBControlClient *cc)
{
  if(cc) {
/*
 * Terminate any extant connection. Note that we can't safely call
 * cc_disconnect_link() here, since that function uses members that
 * might not always be allocated before this destructor function is
 * called.
 */
    int fd = cc_socket_fd(cc);
    if(fd >= 0)
      close(fd);
/*
 * Delete the mutual exclusion guard.
 */
    if(cc->guard_ready) {
      pthread_mutex_destroy(&cc->guard);
      cc->guard_ready = 0;
    };
/*
 * Delete the condition variable.
 */
    if(cc->herald_ready) {
      pthread_cond_destroy(&cc->herald);
      cc->herald_ready = 0;
    };
/*
 * Delete the I/O iterators.
 */
    cc->nss = del_NetSendStr(cc->nss);
    cc->nrs = del_NetReadStr(cc->nrs);
/*
 * Delete the control-command output queue.
 */
    cc->mq = del_CCBMessageQueue(cc->mq);
/*
 * Delete the configuration object.
 */
    cc->cnf = del_CCBConfig(cc->cnf);
/*
 * Finally, delete the container.
 */
    free(cc);
  };
  return NULL;
}

/*.......................................................................
 * Queue a start-scan command to be sent to a given CCB server.
 *
 * Input:
 *  cc  CCBControlClient *  The client side of the control link to the
 *                          CCB.
 *  id              long    An identification number to use to
 *                          refer to this command if the server
 *                          rejects it for some reason.
 *  cnf  const CCBConfig *  The configuration to associate with the
 *                          new scan.
 *  scan   unsigned long    The numeric ID to give the new scan.
 *  mjd         unsigned    This is the date at which the scan should be
 *                          started, expressed in UTC, as a Modified
 *                          Julian Day number.  To be precise, this is
 *                          the integer part of (Julian_Date - 2400000.5).
 *  tod         unsigned    This is the time of day at which the scan
 *                          should be started, specified as the integer
 *                          number of seconds after 0H UTC on the day
 *                          indicated by 'mjd'. The scan starts at the
 *                          start of the specified second, provided that
 *                          the command is received at least one second
 *                          in advance of this time.
 * Output:
 *  return           int    0 - OK.
 *                          1 - Error.
 */
int ccb_cc_queue_start_scan_cmd(CCBControlClient *cc, long id,
				const CCBConfig *cnf, unsigned long scan,
				unsigned long mjd, unsigned long tod)
{
  CCBStartScanCmd msg;   /* The new message */
/*
 * Check the arguments.
 */
  if(!cc) {
    ccb_log_client_msg(cc ? cc->client : NULL, CCB_ERROR, CCB_LOGID(169),
	"ccb_cc_queue_start_scan_cmd: NULL argument(s).\n");
    if(cc) ccb_cc_command_error(cc, id, CCB_CMD_SYSERR);
    return 1;
  };
/*
 * Queue update messages for any parts of the configuration that have changed.
 */
  if(cc_queue_config_updates(cc, id, cnf))
    return 1;
/*
 * Assign the parameters of the command.
 */
  msg.scan = scan;
  msg.mjd = mjd;
  msg.tod = tod;
/*
 * Queue the message for subsequent transmission to the CCB server.
 */
  if(ccb_cc_queue_control_command(cc, &msg.head, CCB_START_SCAN_CMD, id))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a stop-scan command to be sent to a given CCB server.
 *
 * Input:
 *  cc  CCBControlClient *  The client side of the control link to the CCB.
 *  id              long    An identification number to use to
 *                          refer to this command if the server
 *                          rejects it for some reason.
 *  cnf  const CCBConfig *  The configuration to associate with the
 *                          new intra-scan.
 *  scan   unsigned long    The numeric ID to give the new intra-scan.
 * Output:
 *  return           int    0 - OK.
 *                          1 - Error.
 */
int ccb_cc_queue_stop_scan_cmd(CCBControlClient *cc, long id,
			       const CCBConfig *cnf, unsigned long scan)
{
  CCBStopScanCmd msg;   /* The new message */
/*
 * Check the arguments.
 */
  if(!cc) {
    ccb_log_client_msg(cc ? cc->client : NULL, CCB_ERROR, CCB_LOGID(170),
	"ccb_cc_queue_stop_scan_cmd: NULL argument(s).\n");
    if(cc) ccb_cc_command_error(cc, id, CCB_CMD_SYSERR);
    return 1;
  };
/*
 * Queue update messages for any parts of the configuration that have changed.
 */
  if(cc_queue_config_updates(cc, id, cnf))
    return 1;
/*
 * Assign the parameters of the command.
 */
  msg.scan = scan;
/*
 * Queue the message for subsequent transmission to the CCB server.
 */
  if(ccb_cc_queue_control_command(cc, &msg.head, CCB_STOP_SCAN_CMD, id))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a dump-scan command to be sent to a given CCB server.
 *
 * Input:
 *  cc    CCBControlClient *  The client side of the control link to the CCB.
 *  id                long    An identification number to use to
 *                            refer to this command if the server
 *                            rejects it for some reason.
 *  cnf    const CCBConfig *  The configuration to associate with the
 *                            new intra-scan.
 *  scan     unsigned long    The numeric ID to give the new intra-scan.
 *  adc     unsigned short    The ID of the ADC whose samples are to be
 *                            collected (0-15).
 *  samples  unsigned long    The max number of samples to collect per
 *                            integration.
 *  frames   unsigned long    The number of per-integration data-frames
 *                            to deliver.
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error.
 */
int ccb_cc_queue_dump_scan_cmd(CCBControlClient *cc, long id,
			       const CCBConfig *cnf, unsigned long scan,
			       unsigned short adc, unsigned long samples,
			       unsigned long frames)
{
  CCBDumpScanCmd msg;   /* The new message */
/*
 * Check the arguments.
 */
  if(!cc) {
    ccb_log_client_msg(cc ? cc->client : NULL, CCB_ERROR, CCB_LOGID(171),
	"ccb_cc_queue_dump_scan_cmd: NULL argument(s).\n");
    if(cc) ccb_cc_command_error(cc, id, CCB_CMD_SYSERR);
    return 1;
  };
/*
 * Queue update messages for any parts of the configuration that have changed.
 */
  if(cc_queue_config_updates(cc, id, cnf))
    return 1;
/*
 * Assign the parameters of the command.
 */
  msg.scan = scan;
  msg.adc = adc;
  msg.samples = samples;
  msg.frames = frames;
/*
 * Queue the message for subsequent transmission to the CCB server.
 */
  if(ccb_cc_queue_control_command(cc, &msg.head, CCB_DUMP_SCAN_CMD, id))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue configuration command messages for any parts of a given new
 * configuration that differs from a given old configuration.
 *
 * Input:
 *  cc      CCBControlClient *  The client side of the control link to the
 *                              CCB.
 *  id                  long    An identification number to use to
 *                              refer to this command if the server
 *                              rejects it for some reason.
 *  cnf      const CCBConfig *  The new configuration.
 * Output:
 *  return               int    0 - OK.
 *                              1 - Error.
 */
static int cc_queue_config_updates(CCBControlClient *cc, long id,
				   const CCBConfig *cnf)
{
  unsigned diff;         /* The differences between the two configurations */
  CCBControlCommand cmd; /* A configuration command to be queued */
/*
 * Find out what differs between the old and new configurations.
 */
  diff = ccb_diff_config(cnf, cc->cnf);
/*
 * Update the phase-switch configuration if needed.
 */
  if(diff & CCB_CNF_PHASE_SWITCHES) {
    if(ccb_get_phase_switch_cnf(cnf, &cmd.phase_cmd.cnf) ||
       ccb_cc_queue_control_command(cc, &cmd.head, CCB_PHASE_SWITCH_CMD, id))
      return 1;
  };
/*
 * Update the cal-diode configuration if needed.
 */
  if(diff & CCB_CNF_CAL_DIODES) {
    if(ccb_get_cal_diode_cnf(cnf, &cmd.diode_cmd.cnf) ||
       ccb_cc_queue_control_command(cc, &cmd.head, CCB_CAL_DIODE_CMD, id))
      return 1;
  };
/*
 * Update the timing configuration if needed.
 */
  if(diff & CCB_CNF_TIMING) {
    if(ccb_get_timing_cnf(cnf, &cmd.timing_cmd.cnf) ||
       ccb_cc_queue_control_command(cc, &cmd.head, CCB_TIMING_CMD, id))
      return 1;
  };
/*
 * Update the sampler configuration if needed.
 */
  if(diff & CCB_CNF_SAMPLER) {
    if(ccb_get_sampler_cnf(cnf, &cmd.sampler_cmd.cnf) ||
       ccb_cc_queue_control_command(cc, &cmd.head, CCB_SAMPLER_CMD, id))
      return 1;
  };
/*
 * Record the new configuration.
 */
  if(ccb_copy_config(cnf, cc->cnf))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a reset command to be sent to a given CCB server.
 *
 * Input:
 *  cc     CCBControlClient *  The client side of the control link to
 *                             the CCB.
 *  id                 long    An identification number to use to
 *                             refer to this command if the server
 *                             rejects it for some reason.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_cc_queue_reset_cmd(CCBControlClient *cc, long id)
{
  CCBResetCmd msg;   /* The new message */
/*
 * Check the arguments.
 */
  if(!cc) {
    ccb_log_client_msg(cc ? cc->client : NULL, CCB_ERROR, CCB_LOGID(172),
	"ccb_cc_queue_reset_cmd: NULL argument(s).\n");
    if(cc) ccb_cc_command_error(cc, id, CCB_CMD_SYSERR);
    return 1;
  };
/*
 * Queue the message for subsequent transmission to the CCB server.
 */
  if(ccb_cc_queue_control_command(cc, &msg.head, CCB_RESET_CMD, id))
    return 1;
/*
 * Reset the our mirror of the CCB configuration to reflect what the
 * reset command is about to do.
 */
  if(ccb_default_config(cc->cnf))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a ping command to be sent to a given CCB server.
 *
 * Input:
 *  cc     CCBControlClient *  The client side of the control link to
 *                             the CCB.
 *  id                 long    An identification number to use to
 *                             refer to this command if the server
 *                             rejects it for some reason.
 *  seq       unsigned long    The sequence number of the ping command.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_cc_queue_ping_cmd(CCBControlClient *cc, long id, unsigned long seq)
{
  CCBPingCmd msg;   /* The new message */
/*
 * Check the arguments.
 */
  if(!cc) {
    ccb_log_client_msg(cc ? cc->client : NULL, CCB_ERROR, CCB_LOGID(173),
	"ccb_cc_queue_ping_cmd: NULL argument(s).\n");
    if(cc) ccb_cc_command_error(cc, id, CCB_CMD_SYSERR);
    return 1;
  };
/*
 * Fill in the body of the command.
 */
  msg.seq = seq;
/*
 * Queue the message for subsequent transmission to the CCB server.
 */
  if(ccb_cc_queue_control_command(cc, &msg.head, CCB_PING_CMD, id))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a status-request command to be sent to a given CCB server.
 *
 * Input:
 *  cc     CCBControlClient *  The client side of the control link to
 *                             the CCB.
 *  id                 long    An identification number to use to
 *                             refer to this command if the server
 *                             rejects it for some reason.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_cc_queue_status_request_cmd(CCBControlClient *cc, long id)
{
  CCBStatusRequestCmd msg; /* The new message */
/*
 * Check the arguments.
 */
  if(!cc) {
    ccb_log_client_msg(cc ? cc->client : NULL, CCB_ERROR, CCB_LOGID(174),
	"ccb_cc_queue_status_request_cmd: NULL argument(s).\n");
    if(cc) ccb_cc_command_error(cc, id, CCB_CMD_SYSERR);
    return 1;
  };
/*
 * Queue the message for subsequent transmission to the CCB server.
 */
  if(ccb_cc_queue_control_command(cc, &msg.head, CCB_STATUS_REQUEST_CMD, id))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a shutdown command to be sent to a given CCB server.
 *
 * Input:
 *  cc     CCBControlClient *  The client side of the control link to
 *                             the CCB.
 *  id                 long    An identification number to use to
 *                             refer to this command if the server
 *                             rejects it for some reason.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_cc_queue_shutdown_cmd(CCBControlClient *cc, long id)
{
  CCBShutdownCmd msg;   /* The new message */
/*
 * Check the arguments.
 */
  if(!cc) {
    ccb_log_client_msg(cc ? cc->client : NULL, CCB_ERROR, CCB_LOGID(175),
	"ccb_cc_queue_shutdown_cmd: NULL argument(s).\n");
    if(cc) ccb_cc_command_error(cc, id, CCB_CMD_SYSERR);
    return 1;
  };
/*
 * Queue the message for subsequent transmission to the CCB server.
 */
  if(ccb_cc_queue_control_command(cc, &msg.head, CCB_SHUTDOWN_CMD, id))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a reboot command to be sent to a given CCB server.
 *
 * Input:
 *  cc     CCBControlClient *  The client side of the control link to
 *                             the CCB.
 *  id                 long    An identification number to use to
 *                             refer to this command if the server
 *                             rejects it for some reason.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_cc_queue_reboot_cmd(CCBControlClient *cc, long id)
{
  CCBRebootCmd msg;   /* The new message */
/*
 * Check the arguments.
 */
  if(!cc) {
    ccb_log_client_msg(cc ? cc->client : NULL, CCB_ERROR, CCB_LOGID(176),
	"ccb_cc_queue_reboot_cmd: NULL argument(s).\n");
    if(cc) ccb_cc_command_error(cc, id, CCB_CMD_SYSERR);
    return 1;
  };
/*
 * Queue the message for subsequent transmission to the CCB server.
 */
  if(ccb_cc_queue_control_command(cc, &msg.head, CCB_REBOOT_CMD, id))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a monitoring-control command to be sent to a given CCB server.
 *
 * Input:
 *  cc     CCBControlClient *  The client side of the control link to
 *                             the CCB.
 *  id                 long    An identification number to use to
 *                             refer to this command if the server
 *                             rejects it for some reason.
 *  period   unsigned short    The monitoring period, expressed as a
 *                             multiple of the integration period.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_cc_queue_monitor_cmd(CCBControlClient *cc, long id,
			    unsigned short period)
{
  CCBMonitorCmd msg;   /* The new message */
/*
 * Check the arguments.
 */
  if(!cc) {
    ccb_log_client_msg(cc ? cc->client : NULL, CCB_ERROR, CCB_LOGID(177),
	"ccb_cc_queue_monitor_cmd: NULL argument(s).\n");
    if(cc) ccb_cc_command_error(cc, id, CCB_CMD_SYSERR);
    return 1;
  };
/*
 * Assign the parameters of the command.
 */
  msg.period = period;
/*
 * Queue the message for subsequent transmission to the CCB server.
 */
  if(ccb_cc_queue_control_command(cc, &msg.head, CCB_MONITOR_CMD, id))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a telemetry-stream control-command to be sent to a given CCB
 * server.
 *
 * Input:
 *  cc     CCBControlClient *  The client side of the control link to
 *                             the CCB.
 *  id                 long    An identification number to use to
 *                             refer to this command if the server
 *                             rejects it for some reason.
 *  streams  unsigned short    The set of telemetry streams to report to
 *                             the manager, expressed as a bitwise union
 *                             of CCBTelemetryStream enumerators.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_cc_queue_telemetry_cmd(CCBControlClient *cc, long id,
			     unsigned short streams)
{
  CCBTelemetryCmd msg;   /* The new message */
/*
 * Check the arguments.
 */
  if(!cc) {
    ccb_log_client_msg(cc ? cc->client : NULL, CCB_ERROR, CCB_LOGID(178),
	"ccb_cc_queue_telemetry_cmd: NULL argument(s).\n");
    if(cc) ccb_cc_command_error(cc, id, CCB_CMD_SYSERR);
    return 1;
  };
/*
 * Assign the parameters of the command.
 */
  msg.streams = streams;
/*
 * Queue the message for subsequent transmission to the CCB server.
 */
  if(ccb_cc_queue_control_command(cc, &msg.head, CCB_TELEMETRY_CMD, id))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a log-control command to be sent to a given CCB server.
 *
 * Input:
 *  cc     CCBControlClient *  The client side of the control link to
 *                             the CCB.
 *  id                 long    An identification number to use to
 *                             refer to this command if the server
 *                             rejects it for some reason.
 *  period    unsigned long    The time interval at which the
 *                             log-server forgets about historically
 *                             sent messages (seconds).
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_cc_queue_logger_cmd(CCBControlClient *cc, long id,
			    unsigned long period)
{
  CCBLoggerCmd msg;   /* The new message */
/*
 * Check the arguments.
 */
  if(!cc) {
    ccb_log_client_msg(cc ? cc->client : NULL, CCB_ERROR, CCB_LOGID(179),
	"ccb_cc_queue_logger_cmd: NULL argument(s).\n");
    if(cc) ccb_cc_command_error(cc, id, CCB_CMD_SYSERR);
    return 1;
  };
/*
 * Assign the parameters of the command.
 */
  msg.period = period;
/*
 * Queue the message for subsequent transmission to the CCB server.
 */
  if(ccb_cc_queue_control_command(cc, &msg.head, CCB_LOGGER_CMD, id))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a load-driver command to be sent to a given CCB server.
 *
 * Input:
 *  cc     CCBControlClient *  The client side of the control link to
 *                             the CCB.
 *  id                 long    An identification number to use to
 *                             refer to this command if the server
 *                             rejects it for some reason.
 *  type      CCBDriverType    The type of driver to load.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_cc_queue_load_driver_cmd(CCBControlClient *cc, long id,
				 CCBDriverType type)
			    
{
  CCBLoadDriverCmd msg;   /* The new message */
/*
 * Check the arguments.
 */
  if(!cc) {
    ccb_log_client_msg(cc ? cc->client : NULL, CCB_ERROR, CCB_LOGID(180),
	"ccb_cc_queue_load_driver_cmd: NULL argument(s).\n");
    if(cc) ccb_cc_command_error(cc, id, CCB_CMD_SYSERR);
    return 1;
  };
/*
 * Assign the parameters of the command.
 */
  msg.type = type;
/*
 * Queue the message for subsequent transmission to the CCB server.
 */
  if(ccb_cc_queue_control_command(cc, &msg.head, CCB_LOAD_DRIVER_CMD, id))
    return 1;
/*
 * Reset the our mirror of the CCB configuration to reflect what the
 * load-driver command is about to do.
 */
  if(ccb_default_config(cc->cnf))
    return 1;
  return 0;
}

/*.......................................................................
 * Set up the contents of the header of a control command message, then
 * queue it for subsequent transmission to the CCB server.
 *
 * Input:
 *  cc           CCBControlClient *  The client side of the control link
 *                                   to the CCB.
 *  head  CCBControlCommandHeader *  The header to be initilized.
 *  type    CCBControlCommandType    The type of message.
 *  id                       long    The manager-provided ID to use
 *                                   to refer to the command.
 * Output:
 *  return                    int    0 - OK.
 *                                   1 - Error.
 */
static int ccb_cc_queue_control_command(CCBControlClient *cc,
					CCBControlCommandHeader *head,
					CCBControlCommandType type,
					long id)
{
  int status = 0;    /* The return value of cc_queue_control_command() */
/*
 * Check the arguments.
 */
  if(!head) {
    ccb_log_client_msg(cc ? cc->client : NULL, CCB_ERROR, CCB_LOGID(181),
		       "ccb_cc_queue_control_command: NULL argument(s).\n");
    if(cc) ccb_cc_command_error(cc, id, CCB_CMD_SYSERR);
    return 1;
  };
/*
 * Initialize the header of the message.
 */
  head->base.type = type;
  head->id = id;
/*
 * Acquire exclusive access to 'cc' while calling cc_queue_control_command().
 */
  if(cc_lock_client(cc)) {
    ccb_cc_command_error(cc, id, CCB_CMD_SYSERR);
    return 1;
  };
/*
 * Attempt to queue the command (note that cc_queue_control_command())
 * calls ccb_cc_command_error() if there is a problem, so we mustn't
 * redundantly do this here).
 */
  status = cc_queue_control_command(cc, head);
/*
 * Relinquish exclusive access to 'cc'.
 */
  cc_unlock_client(cc);
  return status;
}

/*.......................................................................
 * Queue a command message for transmission to the CCB server over the
 * control link. This involves either adding it to a holding queue, or
 * encoding it for transmission, depending on whether or not the client
 * is currently busy transmitting a previous message.
 *
 * Note that this function requires that the caller acquire exclusive
 * access to the specified CCBControlClient object before calling this
 * function.
 *
 * Input:
 *  cc           CCBControlClient *  The control client.
 *  head  CCBControlCommandHeader *  The header of a new message to queue.
 * Output:
 *  return                    int    0 - OK.
 *                                   1 - Error.
 */
static int cc_queue_control_command(CCBControlClient *cc,
				    CCBControlCommandHeader *head)
{
/*
 * Lookup the size of the message structure that the header belongs to.
 */
  size_t size = ccb_net_native_size(&ccb_control_cmd_table, head->base.type);
  if(!size) {
    cc_command_error(cc, head->id, CCB_CMD_SYSERR);
    return cc_error_while_locked(cc, CCB_ERROR, CCB_LOGID(182),
		  "cc_queue_control_command: Unknown type of message.\n");
  };
/*
 * Append the message to the queue.
 */
  (void) ccb_mq_push(cc->mq, head, size);
/*
 * If the queue was previously empty, and nothing was being sent, the
 * writer will be idle. Get it to queue our message if this is the case.
 */
  if(cc_stage_next_command(cc))
    return 1;
  return 0;
}

/*.......................................................................
 * Unqueue the next queued command to be sent to the manager.
 *
 * Note that the caller must acquire exclusive access to 'cc' before
 * calling this function.
 *
 * Input:
 *  cc       CCBControlClient *  The control client.
 * Input/Output:
 *  msg     CCBControlCommand *  The command message will be recorded
 *                               in *msg.
 * Output:
 *  return                int    0 - OK.
 *                               1 - No message returned.
 */
static int cc_unqueue_next_command(CCBControlClient *cc,
				   CCBControlCommand *msg)
{
  return ccb_mq_pop(cc->mq, msg, sizeof(*msg)) != sizeof(*msg);
}

/*.......................................................................
 * If no message is currently being written to the control socket, get the
 * next message from the output queue and stage it for subsequent
 * transmission to the manager.
 *
 * Note that the caller must acquire exclusive access to 'cc' before
 * calling this function.
 *
 * Input:
 *  cc           CCBControlClient *  The control client.
 *  head  CCBControlCommandHeader *  The header of a particular
 *                                   command message to stage for
 *                                   transmission.
 * Output:
 *  return                    int    0 - OK.
 *                                   1 - Error.
 */
static int cc_stage_next_command(CCBControlClient *cc)
{
  CCBControlCommand cmd;  /* The next command to be queued */
  int status;             /* The return value of a POSIX function */
/*
 * Is the client ready to send a new message?
 */
  if(!cc->write_busy && cc->link_state==CC_LINK_UP) {
/*
 * Get the highest priority message from the output queues.
 */
    if(cc_unqueue_next_command(cc, &cmd))
      return 0;       /* There were no messages waiting to be sent */
/*
 * Encode the message in the network buffer.
 */
    if(ccb_obj_to_net(&ccb_control_cmd_table, nss_get_NetBuf(cc->nss),
		      &cmd.head.base)) {
      cc_command_error(cc, cmd.head.id, CCB_CMD_SYSERR);
      return cc_error_while_locked(cc, CCB_ERROR, CCB_LOGID(183),
	  "cc_stage_command: ccb_obj_to_net() error (%s).\n", strerror(errno));
    };
/*
 * Is either a shutdown or reboot message now being sent?
 */
    switch(cmd.head.base.type) {
    case CCB_SHUTDOWN_CMD:
    case CCB_REBOOT_CMD:
      cc->sending_shutdown = 1;
      break;
    };
/*
 * Record the fact that the writer is now busy transmitting a message.
 */
    cc->write_busy = 1;
/*
 * Tell any waiting writers that there is now something to be written.
 */
    status = pthread_cond_broadcast(&cc->herald);
    if(status) {
      ccb_log_client_msg(cc->client, CCB_ERROR, CCB_LOGID(184),
		 "cc_stage_next_command: pthread_cond_broadcast -> %s\n",
		 strerror(status));
      return 1;
    };
  };
  return 0;
}

/*.......................................................................
 * Query the current status of the control client.
 *
 * Input:
 *  cc           CCBControlClient *  The control client.
 * Output:
 *  return CCBControlClientStatus    The current status of the control
 *                                   client.
 */
CCBControlClientStatus ccb_cc_status(CCBControlClient *cc)
{
/*
 * Assume that the client is down, until proven otherwise.
 */
  int status = CCB_CONTROL_CLIENT_DOWN;
/*
 * Get exclusive access to 'cc' while we examine its state.
 */
  if(cc && cc_lock_client(cc)==0) {
/*
 * Derive the status from the link status and the writer status.
 */
    switch(cc->link_state) {
    case CC_LINK_CONNECTING:
      status = CCB_CONTROL_CLIENT_PEND;
      break;
    case CC_LINK_UP:
      status = cc->write_busy ? CCB_CONTROL_CLIENT_BUSY :
	                        CCB_CONTROL_CLIENT_IDLE;
      break;
    case CC_LINK_DOWN:
    case CC_LINK_DISCONNECTING:
      status = CCB_CONTROL_CLIENT_DOWN;
      break;
    };
/*
 * Relinquish exclusive access to 'cc'.
 */
    cc_unlock_client(cc);
  };
  return status;
}

/*.......................................................................
 * Attempt to send queued commands to the CCB server.
 *
 * Input:
 *  cc   CCBControlClient *  The control client.
 * Output:
 *  return            int    0 - OK.
 *                           1 - Error.
 */
int ccb_cc_send_control(CCBControlClient *cc)
{
  int status;              /* The completion status of cc_send_control() */
/*
 * Check the arguments.
 */
  if(!cc) {
    ccb_log_client_msg(cc ? cc->client : NULL, CCB_ERROR, CCB_LOGID(185),
		       "ccb_cc_send_control: NULL argument(s).\n");
    return 1;
  };
/*
 * Obtain exclusive access to 'cc'.
 */
  if(cc_lock_client(cc))
    return 1;
/*
 * Send pending messages.
 */
  status = cc_send_control(cc);
/*
 * Relinquish exclusive access to 'cc'.
 */
  cc_unlock_client(cc);
  return status;
}

/*.......................................................................
 * Send pending messages to the CCB control server. In blocking I/O mode
 * this function doesn't return until the connection is lost or an
 * error occurs. Otherwise it returns as soon as doing otherwise would
 * cause it to wait.
 *
 * Note that the caller must acquire exclusive access to 'cc' before
 * calling this function.
 *
 * Input:
 *  cc   CCBControlClient *  The control client.
 * Output:
 *  return            int    0 - OK.
 *                           1 - Error.
 */
static int cc_send_control(CCBControlClient *cc)
{
  int write_status;   /* The completion status of nss_send_msg() */
  int status;         /* The return status of pthread functions */
/*
 * Send as many messages as possible.
 */
  do {
/*
 * If blocking I/O has been selected, wait for exclusive write access
 * to the socket when at least one message is waiting to be sent.
 */
    while(cc->blocking && (cc->write_locked || !cc->write_busy)) {
      status = pthread_cond_wait(&cc->herald, &cc->guard);
      if(status) {
	ccb_log_client_msg(cc->client, CCB_ERROR, CCB_LOGID(186),
			   "cc_send_control: pthread_cond_wait -> %s\n",
			   strerror(status));
	return 1;
      };
    };
/*
 * If another thread is currently performing output, defer to it.
 */
    if(cc->write_locked)
      return 0;
/*
 * We can't write anything if the link is either down or being disconnected.
 */
    if(cc->link_state==CC_LINK_DOWN || cc->link_state==CC_LINK_DISCONNECTING) {
      return cc_error_while_locked(cc, CCB_ERROR, CCB_LOGID(187),
	 "Rejected request to send CCB commands while not connected.\n");
/*
 * If a non-blocking connection is being performed, check whether it
 * succeeded.
 */
    } else if(cc->link_state == CC_LINK_CONNECTING &&
	      !cc_connection_completed(cc, ccb_cc_socket(cc))) {
      return 1;
    };
/*
 * Is there anything in the output queue to be written?
 */
    if(cc->write_busy) {
/*
 * Before writing, set the flag that indicates that we have exclusive
 * write access to the control socket, then give up the mutex so that
 * other threads can receive and queue messages while we potentially
 * wait for the write operation to complete.
 */
      cc->write_locked = 1;
      cc_unlock_client(cc);
/*
 * Continue writing the latest command-message to the network.
 */
      write_status = nss_send_msg(cc->nss);
/*
 * Reacquire exclusive access to 'cc' now that we are no longer waiting
 * for the write operation.
 */
      if(cc_lock_client(cc))
	return 1;
/*
 * Relinquish exclusive write access to the control socket.  Note that
 * no writer will actually get a chance to see that we have done this
 * until this function returns and its caller relinquishes access to
 * 'cc'.
 */
      cc->write_locked = 0;
      status = pthread_cond_broadcast(&cc->herald);
      if(status) {
	ccb_log_client_msg(cc->client, CCB_ERROR, CCB_LOGID(188),
			   "cc_send_control: pthread_cond_broadcast -> %s\n",
			   strerror(status));
	return 1;
      };
/*
 * Check the completion status of the write operation.
 */
      switch(write_status) {
      case NET_SEND_DATA:     /* Output temporarily blocked */
	if(!cc->blocking)
	  return 0;
	break;
      case NET_SEND_DONE:     /* Message written, stage the next message */
	cc->write_busy = 0;
/*
 * If we have just finished sending a shutdown or reboot command, disconnect
 * from the server.
 */
	if(cc->sending_shutdown) {
	  ccb_client_disconnect(cc->client);
	  return 1;
	};
/*
 * If there is another queued message to be sent, stage it for transmission.
 */
	if(cc_stage_next_command(cc))
	  return 1;
	break;
      case NET_SEND_CLOSED:
	return cc_error_while_locked(cc, CCB_ERROR, CCB_LOGID(189),
	  "The CCB client attempted to write to a closed control socket.\n");
	break;
      case NET_SEND_ERROR:
      default:
	return cc_error_while_locked(cc, CCB_ERROR, CCB_LOGID(190),
   "The CCB client encountered an I/O error while sending a control reply.\n");
	break;
      };
    };
/*
 * While in blocking I/O mode, return to the start of the loop to
 * await the next message to be sent. In non-blocking I/O mode, we
 * only reach this point if a message was successfully sent without
 * the socket blocking, so if there are any other messages in the output
 * queue, return to the start of the loop to attempt to send the next.
 * Otherwise, if the output queue is empty, exit the loop.
 */
  } while(cc->link_state == CC_LINK_UP && (cc->write_busy || cc->blocking));
  return 0;
}

/*.......................................................................
 * Register the callback function that is to be called whenever a given
 * CCB client object receives a control-status reply message.
 *
 * Input:
 *  cc      CCBControlClient *  The control client.
 *  fn  CCBStatusReplyFn *  The function to call whenever a reply
 *                              to a previously sent control-status
 *                              command is received over the control link.
 *  data                void *  Anonymous data to pass to fn().
 * Output:
 *  return               int    0 - OK.
 *                              1 - Error.
 */
int ccb_cc_status_reply_callback(CCBControlClient *cc,
				      CCBStatusReplyFn *fn, void *data)
{
  if(!cc || !fn) {
    ccb_log_client_msg(cc ? cc->client : NULL, CCB_ERROR, CCB_LOGID(191),
	"ccb_cc_status_reply_callback: NULL argument(s).\n");
    return 1;
  };
/*
 * Install the new callback while we have exclusive access to the cc object.
 */
  if(cc_lock_client(cc))
    return 1;
  cc->cb.status_reply_fn = fn;
  cc->cb.status_reply_data = data;
  cc_unlock_client(cc);
  return 0;
}

/*.......................................................................
 * Register a command-error callback function with a given control-client
 * connection.
 *
 * Input:
 *  cc     CCBControlClient *  The control client.
 *  fn        CCBCmdErrorFn *  The function to call whenever a control
 *                             command evokes an error report.
 *  data               void *  Anonymous data to pass to fn().
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_cc_cmd_error_callback(CCBControlClient *cc, CCBCmdErrorFn *fn,
			      void *data)
{
/*
 * Check the arguments.
 */
  if(!cc || !fn) {
    ccb_log_client_msg(cc ? cc->client : NULL, CCB_ERROR, CCB_LOGID(192),
	"ccb_cc_cmd_error_callback: NULL argument(s).\n");
    return 1;
  };
/*
 * Install the new callback while we have exclusive access to the cc object.
 */
  if(cc_lock_client(cc))
    return 1;
  cc->cb.cmd_error_fn = fn;
  cc->cb.cmd_error_data = data;
  cc_unlock_client(cc);
  return 0;
}

/*.......................................................................
 * Read up to one reply message from the CCB client control socket, and
 * when a complete reply-message has been read, deliver it to the
 * manager by calling the appropriate callback function.
 *
 * Input:
 *  cc    CCBControlClient *  The control client.
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error.
 */
int ccb_cc_read_control(CCBControlClient *cc)
{
  int status;    /* The completion status of cc_read_control() */
/*
 * Check the arguments.
 */
  if(!cc) {
    ccb_log_client_msg(cc ? cc->client : NULL, CCB_ERROR, CCB_LOGID(193),
		"ccb_cc_read_control: NULL argument(s).\n");
    return 1;
  };
/*
 * Obtain exclusive access to 'cc'.
 */
  if(cc_lock_client(cc))
    return 1;
/*
 * Read from the control socket.
 */
  status = cc_read_control(cc);
/*
 * Relinquish exclusive access to 'cc'.
 */
  cc_unlock_client(cc);
  return status;
}

/*.......................................................................
 * Read as many messages from the CCB client control socket as
 * possible, and deliver complete reply-messages to the manager by
 * calling the appropriate callback function. In blocking I/O mode
 * this function doesn't return until the control connection is lost.
 * Otherwise it returns as soon as doing otherwise would cause it to
 * wait.
 *
 * Note that the caller must acquire exclusive access to 'cc' before
 * calling this function.
 *
 * Input:
 *  cc    CCBControlClient *  The control client.
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error.
 */
static int cc_read_control(CCBControlClient *cc)
{
  CCBControlReply reply;   /* A generic reply-message structure */
  int read_status;         /* The return status of nrs_read_msg() */
  int status;              /* The return status of pthread functions */
/*
 * Read as many messages as possible.
 */
  do {
/*
 * If blocking I/O has been selected, wait for exclusive read access
 * to the socket when the link is up.
 */
    while(cc->blocking && cc->read_locked) {
      status = pthread_cond_wait(&cc->herald, &cc->guard);
      if(status) {
	ccb_log_client_msg(cc->client, CCB_ERROR, CCB_LOGID(194),
			   "cc_read_control: pthread_cond_wait -> %s\n",
			   strerror(status));
	return 1;
      };
    };
/*
 * If another thread is currently reading from the control socket, defer to it.
 */
    if(cc->read_locked)
      return 0;
/*
 * We can't read anything if the link is either down, or being disconnected.
 */
    if(cc->link_state==CC_LINK_DOWN || cc->link_state==CC_LINK_DISCONNECTING) {
      return cc_error_while_locked(cc, CCB_ERROR, CCB_LOGID(195),
	 "Rejected request to read CCB replies while not connected.\n");
/*
 * If a non-blocking connection is being performed, check whether it
 * succeeded.
 */
    } else if(cc->link_state == CC_LINK_CONNECTING &&
	      !cc_connection_completed(cc, ccb_cc_socket(cc))) {
      return 1;
    };
/*
 * Before reading, set the flag that indicates that we have exclusive
 * read access to the control socket, then give up the mutex so that
 * other threads can output messages while we potentially wait for the
 * read operation to complete.
 */
    cc->read_locked = 1;
    cc_unlock_client(cc);
/*
 * Attempt to read more data from the control socket.
 */
    read_status = nrs_read_msg(cc->nrs);
/*
 * Reacquire exclusive access to 'cc' now that we are no longer waiting
 * for the read operation.
 */
    if(cc_lock_client(cc))
      return 1;
/*
 * Relinquish exclusive read access to the control socket.  Note that
 * no reader will actually get a chance to see that we have done this
 * until this function returns, and the caller relinquishes access to
 * 'cc'.
 */
    cc->read_locked = 0;
    status = pthread_cond_broadcast(&cc->herald);
    if(status) {
      ccb_log_client_msg(cc->client, CCB_ERROR, CCB_LOGID(196),
			 "cc_read_control: pthread_cond_broadcast -> %s\n",
			 strerror(status));
      return 1;
    };
/*
 * Check the completion status of the read operation.
 */
    switch(read_status) {
    case NET_READ_SIZE:
    case NET_READ_DATA:   /* Input blocked before a message could be read */
      if(!cc->blocking)
	return 0;
      break;
    case NET_READ_DONE:   /* A new message has now been completely read */
/*
 * Decode the message into a message structure.
 */
      if(ccb_net_to_obj(&ccb_control_reply_table, nrs_get_NetBuf(cc->nrs),
		      &reply.head.base)) {
	return cc_error_while_locked(cc, CCB_ERROR, CCB_LOGID(197),
		  "cc_read_control: ccb_net_to_obj() error (%s)",
		  strerror(errno));
      };
/*
 * Invoke the appropriate message-delivery callback.
 */
      if(cc_deliver_reply(cc, &reply))
	return 1;
      break;
    case NET_READ_CLOSED: /* There is no control socket */
      return cc_error_while_locked(cc, CCB_FATAL, CCB_LOGID(198),
        "The CCB server has dropped the control-connection.\n");
      break;
    case NET_READ_ERROR:  /* I/O error */
    default:
      return cc_error_while_locked(cc, CCB_FATAL, CCB_LOGID(199),
         "An I/O error was encountered while reading a CCB control reply.\n");
      break;
    };
/*
 * In blocking I/O mode, keep reading messages until the link is lost,
 * or an error causes the body of the loop to return early. In blocking
 * I/O mode, reaching this point means that we read one message without
 * the socket blocking. In the latter mode, return immediately, rather
 * than attempting to read other messages, so that the caller has the
 * option of servicing other events before calling this function again
 * to read the next message.
 */
  } while(cc->link_state == CC_LINK_UP && cc->blocking);
  return 0;
}

/*.......................................................................
 * Invoke the appropriate callback to deliver a newly received control
 * reply to the manager.
 *
 * Note that the caller must acquire exclusive access to 'cc' before
 * calling this function.
 *
 * Input:
 *  cc    CCBControlClient *  The control client.
 *  reply  CCBControlReply *  The reply to be delivered.
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error.
 */
static int cc_deliver_reply(CCBControlClient *cc, CCBControlReply *reply)
{
  int no_callback = 0;   /* True if a callback is missing */
  int waserr = 0;        /* True if an error occurs */
/*
 * Get a snapshot of the container of the callbacks while we have
 * exclusive access to it.
 */
  CCBControlClientCallbacks cb = cc->cb;
/*
 * To avoid the risk of a callback recursively locking 'cc' by calling
 * another function in this module, temporarily relinquish exclusive
 * access to 'cc' while we call the manager's callback.
 */
  cc_unlock_client(cc);
/*
 * Locate the callback associated with the type of reply that has been
 * provided.
 */
  switch(reply->head.base.type) {
  case CCB_CNTRL_PING_REPLY:           /* A reply to a ping command */
    waserr = ccb_report_ping_reply(cc->client, CCB_CNTRL_LINK,
				   reply->ping.seq);
    break;
  case CCB_STATUS_REPLY:         /* A reply to a status-request command */
    if(cb.status_reply_fn) {
      CCBStatusReply *msg = &reply->status;
      waserr = cb.status_reply_fn(cc->client,
					cb.status_reply_data,
					msg->status);
    } else {
      no_callback = 1;
    };
    break;
  case CCB_CNTRL_CMD_ACK:           /* A command-acknowledgment reply */
    {
      CCBCntrlCmdAck *msg = &reply->cmd_ack;
      if(msg->status != CCB_CMD_ACCEPTED)
	waserr = ccb_cc_command_error(cc, msg->id, msg->status);
    };
    break;
  default:
    ccb_log_client_msg(cc->client, CCB_ERROR, CCB_LOGID(200),
      "cc_deliver_reply: Can't deliver unknown type of reply.\n");
    waserr = 1;
    break;
  };
/*
 * Reacquire exclusive access to 'cc'.
 */
  if(cc_lock_client(cc))
    return 1;
/*
 * If the reply type was known by the above switch statement, but
 * the client process hadn't registered a callback for it, report
 * this here.
 */
  if(no_callback) {
    waserr = cc_error_while_locked(cc, CCB_ERROR, CCB_LOGID(201),
		"No callback has been provided to deliver \"%s\" replies.\n",
		ccb_net_obj_name(&ccb_control_reply_table,
				 reply->head.base.type));
  };
  return waserr;
}

/*.......................................................................
 * Report an error regarding a CCB control-command of a given ID.
 *
 * Input:
 *  cc   CCBControlClient *  The networking interface of the control
 *                           client.
 *  id               long    The manager-specified identifier of the
 *                           control-command that suffered the error.
 *  status   CCBCmdStatus    The completion status of the offending
 *                           command.
 * Output:
 *  return            int    0 - OK.
 *                           1 - Error.
 */
int ccb_cc_command_error(CCBControlClient *cc, long id, CCBCmdStatus status)
{
  int retval;  /* The return value of cc_command_error() */
/*
 * Check the arguments.
 */
  if(!cc) {
    ccb_log_client_msg(cc ? cc->client : NULL, CCB_ERROR, CCB_LOGID(202),
	"ccb_cc_command_error: NULL argument(s).\n");
    return 1;
  };
/*
 * Acquire exclusive access to 'cc'.
 */
  if(cc_lock_client(cc))
    return 1;
/*
 * Call the manager's error callback.
 */
  retval = cc_command_error(cc, id, status);
/*
 * Relinquish exclusive access to 'cc'.
 */
  cc_unlock_client(cc);
  return retval;
}

/*.......................................................................
 * Report an error regarding a CCB control-command of a given ID.
 *
 * Note that the caller must acquire exclusive access to 'cc' before
 * calling this function.
 *
 * Input:
 *  cc   CCBControlClient *  The networking interface of the control
 *                           client.
 *  id               long    The manager-specified identifier of the
 *                           control-command that suffered the error.
 *  status   CCBCmdStatus    The completion status of the offending
 *                           command.
 * Output:
 *  return            int    0 - OK.
 *                           1 - Error.
 */
static int cc_command_error(CCBControlClient *cc, long id, CCBCmdStatus status)
{
  int retval = 0;   /* The return value of this function */
/*
 * Lookup the command-error callback function while we still have
 * exclusive access to 'cc'.
 */
  CCBCmdErrorFn *fn = cc->cb.cmd_error_fn;
  void *data = cc->cb.cmd_error_data;
/*
 * Has the manager expressed interest in being notified of command errors?
 */
  if(fn) {
/*
 * To avoid the risk of a callback recursively locking 'cc' by calling
 * another function in this module, temporarily relinquish exclusive
 * access to 'cc' while we call the manager's callback.
 */
    cc_unlock_client(cc);
/*
 * Use the manager's callback to deliver the error.
 */
    retval = fn(cc->client, data, id, status);
/*
 * Reacquire exclusive access to 'cc'.
 */
    if(cc_lock_client(cc))
      return 1;
  };
  return retval;
}

/*.......................................................................
 * Return the file descriptor of the current control-link socket, if
 * any.
 *
 * Input:
 *  cc   CCBControlClient *  The networking interface of the control
 *                           client.
 * Output:
 *  return            int    The socket file descriptor, or -1 if
 *                           not currently connected to a CCB control
 *                           server.
 */
int ccb_cc_socket(CCBControlClient *cc)
{
  int fd;   /* The file descriptor of the socket */
/*
 * Check the arguments.
 */
  if(!cc)
    return -1;
/*
 * Acquire exclusive access to the cc object.
 */
  if(cc_lock_client(cc))
    return -1;
/*
 * Get the file descriptor of the socket.
 */
  fd = cc_socket_fd(cc);
/*
 * Relinquish exclusive access to the cc object.
 */
  cc_unlock_client(cc);
  return fd;
}

/*.......................................................................
 * Return the file-descriptor of a given CCBControlClient object. Note
 * that this is an internal function which assumes that its caller has
 * acquired the guard mutex of the object.
 *
 * Input:
 *  cc   CCBControlClient *  The networking interface of the control
 *                           client.
 * Output:
 *  return            int    The socket file descriptor, or -1 if
 *                           not currently connected to a CCB control
 *                           server.
 */
static int cc_socket_fd(CCBControlClient *cc)
{
  return cc ? nrs_get_fd(cc->nrs) : -1;
}

/*.......................................................................
 * Initiate a connection to a CCB control server.
 *
 * Input:
 *  cc  CCBControlClient *  The client side of the CCB control link.
 *  host      const char *  The address of the CCB server to contact.
 *                          When the 'nonblocking' argument is zero,
 *                          this can be either a numeric IP address or
 *                          a host name. When the 'nonblocking' argument
 *                          is non-zero, only numeric IP addresses are
 *                          accepted, since name-server lookups take
 *                          time.
 *  nonblocking      int    When this argument is 0, this function
 *                          doesn't return until the control link has
 *                          either been established or rejected. When
 *                          non-zero, a non-blocking connect is
 *                          initiated to the CCB control server, and
 *                          this function may then return before
 *                          this connection has completed. When this
 *                          happens, responding to the completion or
 *                          rejection of the connection is left to
 *                          a subsequent call to ccb_cc_send_control()
 *                          and/or ccb_cc_read_control() depending on
 *                          which type of I/O the completion elicits.
 * Output:
 *  return           int    0 - OK.
 *                          1 - Error.
 */
int ccb_cc_connect(CCBControlClient *cc, const char *host, int nonblocking)
{
  int status=0;   /* The return status while holding the client lock */
/*
 * Check the arguments.
 */
  if(!cc || !host) {
    ccb_log_client_msg(cc ? cc->client : NULL, CCB_ERROR, CCB_LOGID(203),
	"ccb_cc_connect: NULL argument(s).\n");
    return 1;
  };
/*
 * Get exclusive access to the cc object.
 */
  if(cc_lock_client(cc))
    return 1;
/*
 * Attempt to connect to the control server.
 */
  status = cc_connect_link(cc, host, nonblocking);
/*
 * Relinquish exclusive access to 'cc'.
 */
  cc_unlock_client(cc);
  return status;
}

/*.......................................................................
 * Initiate a connection to a CCB control server.
 *
 * Note that the caller must acquire exclusive access to 'cc' before
 * calling this function.
 *
 * Input:
 *  cc  CCBControlClient *  The client side of the CCB control link.
 *  host      const char *  The address of the CCB server to contact.
 *                          When the 'nonblocking' argument is zero,
 *                          this can be either a numeric IP address or
 *                          a host name. When the 'nonblocking' argument
 *                          is non-zero, only numeric IP addresses are
 *                          accepted, since name-server lookups take
 *                          time.
 *  nonblocking      int    When this argument is 0, this function
 *                          doesn't return until the control link has
 *                          either been established or rejected. When
 *                          non-zero, a non-blocking connect is
 *                          initiated to the CCB control server, and
 *                          this function may then return before
 *                          this connection has completed. When this
 *                          happens, responding to the completion or
 *                          rejection of the connection is left to
 *                          a subsequent call to ccb_cc_send_control()
 *                          and/or ccb_cc_read_control() depending on
 *                          which type of I/O the completion elicits.
 * Output:
 *  return           int    0 - OK.
 *                          1 - Error.
 */
static int cc_connect_link(CCBControlClient *cc, const char *host,
			   int nonblocking)
{
  int fd;         /* The file descriptor of the new socket */
  int pending;    /* True if a non-blocking connection is pending completion */
/*
 * Terminate any existing control connection.
 */
  cc_disconnect_link(cc);
/*
 * Mark the link as being in the process of being connected.
 */
  cc->link_state = CC_LINK_CONNECTING;
/*
 * Install the default power-on CCB configuration.
 */
  if(ccb_default_config(cc->cnf))
    return 1;
/*
 * If a blocking connect is being used, or a name-server lookup being
 * performed, tcp_connect() may block indefinitely, so for now
 * release exclusive access to the cc object, so that other threads
 * can queue messages to be sent when the connection eventually
 * completes.
 */
  cc_unlock_client(cc);
/*
 * Attempt to connect to the CCB control-port on the specified host.
 */
  errno = 0;
  fd = tcp_connect(host, CCB_CONTROL_PORT, !nonblocking);
/*
 * Is a non-blocking connect still in progress?
 */
  pending = nonblocking && errno == EINPROGRESS;
/*
 * Reacquire exclusive access to the cc object. If this fails,
 * cancel the new connection.
 */
  if(cc_lock_client(cc)) {
    cc->link_state = CC_LINK_DOWN;
    if(fd >= 0)
      close(fd);
    return 1;
  };
/*
 * Connection failed?
 */
  if(fd < 0) {
    cc->link_state = CC_LINK_DOWN;
    cc_error_while_locked(cc, CCB_ERROR, CCB_LOGID(204),
		       "Couldn't connect to CCB control-server at %s (%s).\n",
		       host, strerror(errno));
    return 1;
  };
/*
 * Did another thread request a disconnect or another connection while
 * we didn't have exclusive access to 'cc'?
 */
  if(cc->link_state != CC_LINK_CONNECTING) {
    close(fd);
  } else {
/*
 * Attach the network input and output streams to the new socket.
 */
    attach_NetReadStr(cc->nrs, fd);
    attach_NetSendStr(cc->nss, fd);
/*
 * Record the non-blocking state of the socket.
 */
    cc->blocking = !nonblocking;
/*
 * If the connection completed successfully, attempt to activate the
 * control link.
 */
    if(!pending && cc_activate_link(cc)) {
      cc_disconnect_link(cc);
      return 1;
    };
  };
  return 0;
}

/*.......................................................................
 * Disconnect from a remote CCB control server.
 *
 * Input:
 *  cc  CCBControlClient *  The client side of the CCB control link.
 */
void ccb_cc_disconnect(CCBControlClient *cc)
{
  if(cc) {
/*
 * Get exclusive access to the cc object.
 */
    if(cc_lock_client(cc))
      return;
/*
 * Disconnect the link.
 */
    cc_disconnect_link(cc);
/*
 * Relinquish exclusive access to the cc object.
 */
    cc_unlock_client(cc);
  };
  return;
}

/*.......................................................................
 * Disconnect from a remote CCB control server. This is an internal
 * function which must only be called by functions that have already
 * acquired exclusive access to the associated CCBControlClient object.
 *
 * Input:
 *  cc  CCBControlClient *  The client side of the CCB control link.
 */
static void cc_disconnect_link(CCBControlClient *cc)
{
/*
 * Is there a link to be terminated?
 */
  if(cc->link_state != CC_LINK_DOWN &&
     cc->link_state != CC_LINK_DISCONNECTING) {
    int fd;  /* The file descriptor of the socket */
/*
 * Record the fact that the link is in the process of being disconnected.
 */
    cc->link_state = CC_LINK_DISCONNECTING;
/*
 * Get the file-descriptor of the control socket.
 */
    fd = cc_socket_fd(cc);
/*
 * Dettach the socket from the I/O iterators.
 */
    detach_NetSendStr(cc->nss);
    detach_NetReadStr(cc->nrs);
/*
 * Close the control socket.
 */
    if(fd >= 0)
      close(fd);
  };
/*
 * Discard any pending output.
 */
  cc->write_busy = 0;
  ccb_mq_purge(cc->mq);
/*
 * No shutdown or reboot message is being sent.
 */
  cc->sending_shutdown = 0;
/*
 * Mark the control connection as down.
 */
  cc->link_state = CC_LINK_DOWN;
}

/*.......................................................................
 * When performing a non-blocking connect() on a socket, when select()
 * says that the socket has become either readable or writable (or
 * both), this function is called to ascertain whether the connection
 * succeeded or not. Note that before calling this function, the caller
 * must acquire exclusive access to 'cc'.
 *
 * Input:
 *  cc  CCBControlClient *  The client side of the CCB control link.
 *  sock             int    The socket to check, to see if a non-blocking
 *                          connection succeeded or failed.
 * Output:
 *  return           int    1 - The connection completed successfully.
 *                          0 - The connection failed.
 */
static int cc_connection_completed(CCBControlClient *cc, int sock)
{
  int status;                   /* The error status returned by getsockopt() */
  socklen_t status_len = sizeof(status); /* The size of 'status' */
/*
 * Query the socket's error flag. Note that whereas most implementations return
 * 0, and leave the error status in the 'status' argument, under Solaris
 * getsockopt() returns -1, and the error is left in errno.
 */
  if(getsockopt(sock, SOL_SOCKET, SO_ERROR, &status, &status_len) < 0)
    status = errno;
/*
 * Connection failed?
 */
  if(status) {
    (void) cc_error_while_locked(cc, CCB_ERROR, CCB_LOGID(205),
		       "Couldn't connect to CCB control server (%s).\n",
		       strerror(status));
    return 0;
  };
/*
 * Activate the link.
 */
  if(cc_activate_link(cc))
    return 0; /* Failed */
/*
 * The connection succeeded.
 */
  return 1;
}

/*.......................................................................
 * When a new connection is established the following function is called
 * to activate the link. Note that this is an internal function, and
 * it must only be called after the caller has acquired exclusive access
 * to the specified cc object. Note that before calling this function,
 * the caller must acquire exclusive access to 'cc'.
 *
 * Input:
 *  cc  CCBControlClient *  The client side of the CCB control link.
 * Output:
 *  return           int    0 - OK.
 *                          1 - Error.
 */
static int cc_activate_link(CCBControlClient *cc)
{
  int status;  /* The return status of pthread functions */
/*
 * Mark the link as up.
 */
  cc->link_state = CC_LINK_UP;
/*
 * Tell any waiting readers that they can now procede to call read().
 */
  status = pthread_cond_broadcast(&cc->herald);
  if(status) {
    ccb_log_client_msg(cc->client, CCB_ERROR, CCB_LOGID(206),
		       "cc_activate_link: pthread_cond_broadcast -> %s\n",
		       strerror(status));
    return 1;
  };
/*
 * If there are any commands in the output queue, stage the first of
 * them for transmission.
 */
  if(cc_stage_next_command(cc))
    return 1;
  return 0;
}

/*.......................................................................
 * Acquire exclusive access to a given CCBControlClient object.
 *
 * Input:
 *  cc  CCBControlClient *  The client side of the CCB control link.
 * Output:
 *  return           int    0 - Exclusive access granted.
 *                          1 - Couldn't get access.
 */
static int cc_lock_client(CCBControlClient *cc)
{
  int status = pthread_mutex_lock(&cc->guard);
  if(status) {
    ccb_log_client_msg(cc->client, CCB_ERROR, CCB_LOGID(207),
		       "Couldn't get control-client lock.\n");
    return 1;
  }
  return 0;
}

/*.......................................................................
 * Relinquish exclusive access to a given CCBControlClient object.
 *
 * Input:
 *  cc  CCBControlClient *  The client side of the CCB control link.
 * Output:
 *  return           int    0 - Access relinquished.
 *                          1 - Error.
 */
static int cc_unlock_client(CCBControlClient *cc)
{
  int status = pthread_mutex_unlock(&cc->guard);
/*
 * If unlocking fails we can't report the error via the manager's callback,
 * since we still have the mutex locked, so pass NULL instead of cc->client,
 * to tell ccb_log_client_msg() to write to stderr.
 */
  if(status) {
    ccb_log_client_msg(NULL, CCB_ERROR, CCB_LOGID(208),
		       "Couldn't release control-client lock.\n");
    return 1;
  }
  return 0;
}

/*.......................................................................
 * This function is used to safely report errors when the caller is
 * holding the mutex guard of a CCBControlClient object. The return value
 * is 1, which is the error-return values of most functions in this file,
 * so it can be used as the expression of return statements.
 *
 * Input:
 *  cc  CCBControlClient *  The client side of the CCB control link.
 *  level    CCBLogLevel    The significance of the message.
 *  id     unsigned long    The identifier of the calling statement.
 *  fmt       const char *  The printf-style format string of the
 *                          message.
 *  ...                     Any arguments referred to by the format.
 * Output:
 *  return           int    Always 1.
 */
static int cc_error_while_locked(CCBControlClient *cc, CCBLogLevel level,
				 unsigned long id, const char *fmt, ...)
{
  va_list ap;               /* The encapsulated variable argument list */
/*
 * Relinquish exclusive access to 'cc' while reporting the error.
 * This allows the manager's callback function to be invoked without
 * fear that it might call a function in the library that attempted
 * to recursively acquire the guard mutex (which would hang the program).
 */
  cc_unlock_client(cc);
/*
 * Format the error message, then deliver it to the manager.
 */
  va_start(ap, fmt);
  ccb_vlog_client_msg(cc->client, level, id, fmt, ap);
  va_end(ap);
/*
 * Reacquire exclusive access to 'cc', since the caller is expecting
 * to still have it when we return.
 */
  (void) cc_lock_client(cc);
/*
 * Return the value that most functions in this file usually return
 * on error.
 */
  return 1;
}

/*.......................................................................
 * Select between blocking and non-blocking control-client socket I/O.
 * Input:
 *  cc   CCBControlClient *  The networking interface of the control
 *                           client.
 *  on                int    True to select non-blocking I/O.
 *                           False to select blocking I/O.
 * Output:
 *  return            int    0 - OK.
 *                           1 - Error.
 */
int ccb_cc_non_blocking_io(CCBControlClient *cc, int on)
{
  int fd;   /* The file-descriptor of the current control socket */
/*
 * Check the arguments.
 */
  if(!cc) {
    ccb_log_client_msg(cc ? cc->client : NULL, CCB_ERROR, CCB_LOGID(209),
	"ccb_cc_non_blocking_io: NULL argument(s).\n");
    return 1;
  };
/*
 * Get the file-descriptor of the control socket.
 */
  fd = cc_socket_fd(cc);
/*
 * Establish the requested form of I/O.
 */
  if(fd >= 0 && tcp_set_blocking(fd, !on)) {
    ccb_log_client_msg(cc->client, CCB_ERROR, CCB_LOGID(210),
	"Error establishing %sblocking socket I/O.", on ? "non-":"");    
    return 1;
  };
/*
 * Record the new state.
 */
  cc->blocking = !on;
  return 0;
}
