#ifndef ccbnetobj_h
#define ccbnetobj_h

#include <stddef.h>  /* size_t */

#include "netbuf.h"

typedef struct CCBNetMsgTable CCBNetMsgTable;

/*
 * Check the contents of a table of message descriptions.
 */
int ccb_check_net_table(const CCBNetMsgTable *types, char *errbuf,
			size_t errdim);

/* Enumerate known basic database types */

typedef enum {
  CCB_NET_MSG,     /* The CCBNetMsg structure that is required to be the */
                   /*  first member of all message structures. */
  CCB_NET_ASCII,   /* Elements of a C string cast to unsigned 8-bit integers */
  CCB_NET_BYTE,    /* 8-bit signed integers */
  CCB_NET_UBYTE,   /* 8-bit unsigned integers */
  CCB_NET_SHORT,   /* 16-bit signed integers */
  CCB_NET_USHORT,  /* 16-bit unsigned integers */
  CCB_NET_LONG,    /* 32-bit signed integers */
  CCB_NET_ULONG,   /* 32-bit unsigned integers */
  CCB_NET_FLOAT,   /* 32-bit floating point */
  CCB_NET_DOUBLE   /* 64-bit floating point */
} CCBNetDataType;

/*
 * All CCB network messages start with a member of the following type.
 */
typedef struct {
  long type;  /* The type of the parent message-structure */
} CCBNetMsg;

/*
 * Pack the contents of a network message structure into a network buffer.
 */
int ccb_obj_to_net(const CCBNetMsgTable *types, NetBuf *net, CCBNetMsg *msg);

/*
 * Unpack the contents of a network buffer into the corresponding
 * network message structure.
 */
int ccb_net_to_obj(const CCBNetMsgTable *types, NetBuf *net, CCBNetMsg *msg);

/* Declare a type to be used to describe one member of a message structure */

typedef struct {
  const char *name;    /* Member name */
  size_t offset;       /* Byte offset of member within the message structure */
  CCBNetDataType type; /* The data-type of the member */
  int ntype;           /* The number of elements of type 'type' in the member */
} CCBNetMsgMember;

/* Declare an object description type */

typedef struct {
  long type;                     /* The enumerated type of the message */
  const char *name;              /* Name to refer to object by */
  const CCBNetMsgMember *member; /* An array of descriptions of the 'nmember' */
                                 /*  members in the message structure. */
  int nmember;                   /* The number of elements in member[] */
  size_t native_size;            /* The host-dependent size of the message */
                                 /*  structure. */
} CCBNetMsgInfo;

/*
 * A container of the following type should be used to describe each
 * of the network objects that are to be supported.
 */
struct CCBNetMsgTable {
  const char *name;          /* The name of the object collection */
  const CCBNetMsgInfo *info; /* Descriptions of each of nobj objects */
                             /* This can be NULL if nobj==0. */
  int nmsg;                  /* The number of message structures in the */
                             /*  table. */
  size_t union_size;         /* The size of a union of all message structures */
};

size_t ccb_net_max_obj_size(const CCBNetMsgTable *types);
size_t ccb_net_obj_size(const CCBNetMsgTable *types, int id);
size_t ccb_net_type_size(CCBNetDataType type);
size_t ccb_net_native_size(const CCBNetMsgTable *types, int id);
const char *ccb_net_obj_name(const CCBNetMsgTable *types, int id);
const CCBNetMsgInfo *ccb_net_obj_info(const CCBNetMsgTable *types, int id);

#endif
